"""Common device class for iNels BUS."""
import logging

from enum import Enum

_LOGGER = logging.getLogger(__name__)


class DeviceType(Enum):
    """Enums for InelsDevice type"""
    UNDEFINED = 'undefined'
    LIGHT = 'lights'
    SWITCH = 'on_off'
    SHUTTER = 'shutters'
    THERM = 'heat-control'
    HEATING = 'heating'
    GARAGE = 'garage'

    @staticmethod
    def is_in(item):
        """Test when the item is in enum by value."""
        types = set(item.value for item in DeviceType)
        # when the item is in the enum then return enum object, otherwise
        # return object type UNDEFINED
        if item not in types:
            return DeviceType.UNDEFINED
        else:
            return DeviceType(item)


class InelsDevice:
    """Device class."""

    def __init__(self, title, id, type, proxy):
        """Initialize InelsDevice class."""
        self.title = title
        self.id = id
        self.type = type
        self.proxy = proxy
        self.temp_current: str = None
        self.rele: str = None
        self.temp_set: str = None
        self.down: str = None
        self.up: str = None
        self.read_only: bool = False
        self.value = None

    def loadFromJson(self, json_obj):
        """Load addition data from json file except id, type and title."""
        if 'therm' in json_obj:
            self.temp_current = json_obj['therm']

        if 'rele' in json_obj:
            self.rele = json_obj['rele']

        if 'stateth' in json_obj:
            self.temp_set = json_obj['stateth']

        if 'down' in json_obj:
            self.down = json_obj['down']

        if 'up' in json_obj:
            self.up = json_obj['up']

        if 'read_only' in json_obj:
            self.read_only = json_obj['read_only'] != 'no'

    def observe(self):
        """Read the current value of the device."""
        raw = self.proxy.read([self.id])
        value = raw[self.id]
        self.value = value

        return value

    def set_value(self, value):
        """Set value to the device."""
        curr_int = isinstance(self.value, int)
        new_int = isinstance(value, int)

        if (self.value != value or curr_int is not new_int):
            self.value = value
            self._write()

    def _write(self):
        """Write data to the iNels BUS unit."""
        self.proxy.write(self, self.value)

    @property
    def is_available(self):
        # first test when device has any value
        if self.value != None:
            return True
        else:
            # if not then try to observer
            self.observe()
        # when nothing change then device is not available
        return self.value != None



"""Unit testing of iNels BUS CU3 library."""

from tests.const_test import (
    TEST_HOST,
    TEST_INELS_BUS3_NAMESPACE,
    TEST_PORT,
    TEST_RAW_DEVICES,
    TEST_ROOMS,
    TEST_DATA_GARAGE,
    TEST_DATA_HEATING,
    TEST_DATA_LIGHT,
    TEST_DATA_SHUTTER,
    TEST_DATA_SWITCH,
    TEST_DATA_THERM
)
from inels.cu3 import (
    InelsBus3,
    InelsBusException,
    InelsBusDataTypeException
)
from inels.device import DeviceType, InelsDevice
from unittest.mock import patch, Mock
from unittest import TestCase


class InelsBus3Test(TestCase):
    """Class to test iNels BUS CU3 library."""

    @patch(TEST_INELS_BUS3_NAMESPACE)
    def test_class_calling(self, mock_class):
        """Class instance test."""
        mock_class(TEST_HOST, TEST_PORT)

        mock_class.assert_called()
        mock_class.assert_called_once()
        mock_class.assert_called_with(TEST_HOST, TEST_PORT)

        self.assertEqual(mock_class.call_count, 1)

    @patch('xmlrpc.client.ServerProxy')
    def test_connection_failed(self, mock_server):
        """Test proxy connection."""
        mock_server.return_value = Mock()
        mock_server.side_effect = InelsBusException(
            'common_exception', 'Exception occur')

        inels = InelsBus3(TEST_HOST, TEST_PORT)

        with patch.object(inels, 'ping', return_value=True):
            ret = inels.ping()
            self.assertEqual(True, ret)

        self.assertEqual(mock_server.call_count, 0)

    @patch(f'{TEST_INELS_BUS3_NAMESPACE}.ping')
    def test_ping_failed(self, mock_method):
        """Test ping method."""
        mock_method.return_value = False

        ret = InelsBus3(TEST_HOST, TEST_PORT).ping()

        self.assertEqual(mock_method.call_count, 1)
        self.assertFalse(ret)

    @patch(f'{TEST_INELS_BUS3_NAMESPACE}.getPlcIp')
    def test_getPlcIp_success(self, mock_method):
        """Test Ip address of the PLC."""
        mock_method.return_value = "192.168.2.10"

        ret = InelsBus3(TEST_HOST, TEST_PORT).getPlcIp()

        self.assertEqual(mock_method.call_count, 1)
        self.assertEqual(ret, "192.168.2.10")

    @patch(f'{TEST_INELS_BUS3_NAMESPACE}.getRooms')
    def test_getRoomsRaw_list(self, mock_method):
        """Test list of rooms defined on Connection server."""
        mock_method.return_value = TEST_ROOMS

        ret = InelsBus3(TEST_HOST, TEST_PORT).getRooms()

        self.assertEqual(mock_method.call_count, 1)
        self.assertEqual(len(ret), 6)
        self.assertEqual(ret[1], 'First floor')

    @patch(f'{TEST_INELS_BUS3_NAMESPACE}.getRoomDevicesRaw')
    def test_getRoomDevices_list(self, mock_method):
        """Test list of all devices in room."""
        mock_method.return_value = TEST_RAW_DEVICES

        inels = InelsBus3(TEST_HOST, TEST_PORT)
        raw = inels.getRoomDevicesRaw('room')

        self.assertEqual(mock_method.call_count, 1)
        self.assertEqual(len(raw), len(mock_method.return_value))

        # this is the way how to mock private method
        with patch.object(
            inels,
            '_InelsBus3__readDeviceData',
            return_value={'Doors_Garage': 0}
        ):
            with patch.object(inels, 'ping', return_value=True):
                json_obj = inels.getRoomDevices('room')

                device = json_obj[0]

                self.assertGreater(len(json_obj), 0)
                self.assertEqual(device.id, 'Doors_Garage')
                self.assertEqual(device.title, 'Doors')

                with patch.object(inels, 'read',
                                  return_value="{'Doors_Garage': '0'}"):
                    device_value = inels.read(device.id)
                    self.assertEqual(device_value, "{'Doors_Garage': '0'}")

    @patch(f'{TEST_INELS_BUS3_NAMESPACE}.getRooms')
    def test_getAllDevices(self, mock_method):
        """Test to load all devices in all rooms.
        We have simulation only for one room."""
        mock_method.return_value = ['Garage']

        inels = InelsBus3(TEST_HOST, TEST_PORT)

        with patch.object(
            inels,
            'getRoomDevicesRaw',
            return_value=TEST_RAW_DEVICES
        ):
            with patch.object(
                inels,
                '_InelsBus3__readDeviceData',
                return_value={'Doors_Garage': 0}
            ):
                with patch.object(inels, 'ping', return_value=True):
                    devices = inels.getAllDevices()
                    self.assertGreater(len(devices), 0)

                    lights = [x for x in devices if x.type
                              is DeviceType.LIGHT]
                    self.assertEqual(len(lights), 2)
                    garage = [x for x in devices if x.type
                              is DeviceType.GARAGE]
                    self.assertEqual(len(garage), 1)

    def test_device_type_is_in(self):
        """Test device type in inels device."""
        d_type = DeviceType.LIGHT
        d_type_res = DeviceType.is_in(d_type.value)

        self.assertEqual(d_type, d_type_res)

    def test_device_type_is_not_in(self):
        """Test device type which is not in enum."""
        d_type = "undefined_type"
        d_type_res = DeviceType.is_in(d_type)

        self.assertNotEqual(d_type, d_type_res)
        self.assertEqual(d_type_res, DeviceType.UNDEFINED)

    def test_read_bad_request(self):
        """Test observing data from device with bad response."""
        inels = InelsBus3(TEST_HOST, TEST_PORT)

        with patch.object(
            inels,
            '_InelsBus3__readDeviceData',
            return_value={'Garage_door': 0}
        ) as read:
            read.side_effect = InelsBusDataTypeException(
                "error", "bad data type")
            with self.assertRaises(InelsBusDataTypeException):
                data = inels.read('id')
                read.assert_called_once_with()
                self.assertIsNone(data)

    def test_observe_request_success(self):
        """Test observing data from devices."""
        inels = InelsBus3(TEST_HOST, TEST_PORT)

        with patch.object(
            inels,
            '_InelsBus3__readDeviceData',
            return_value={'Garage_door': 0}
        ):
            data = inels.read(['Garage_door'])
            self.assertIsNotNone(data)
            self.assertEqual(data['Garage_door'], 0)

    def test_write_bad_request(self):
        """Test write data to the proxy with bad request."""
        inels = InelsBus3(TEST_HOST, TEST_PORT)

        with patch.object(
            inels,
            '_InelsBus3__writeValues',
            return_value=None
        ) as write:
            write.side_effect = InelsBusDataTypeException(
                'write', 'Bad data type to write.')

            with self.assertRaises(InelsBusDataTypeException) as exc:
                with patch.object(inels, 'ping', return_value=True):
                    inels.write("", "")
                    exc.assert_called_once_with()
                    self.assertEqual(write.call_count, 0)

    @patch(f'{TEST_INELS_BUS3_NAMESPACE}.getRooms')
    def test_write_request_success(self, mock_method):
        """Test to write data to the proxy."""
        mock_method.return_value = ['Garage']

        inels = InelsBus3(TEST_HOST, TEST_PORT)

        with patch.object(
            inels,
            'getRoomDevicesRaw',
            return_value=TEST_RAW_DEVICES
        ):
            with patch.object(
                inels,
                '_InelsBus3__readDeviceData',
                return_value={'Doors_Garage': 0}
            ):
                with patch.object(inels, 'ping', return_value=True):
                    devices = inels.getAllDevices()
                    self.assertEqual(len(devices), 3)

                    lights = [x for x in devices if x.type is DeviceType.LIGHT]

                    with patch.object(
                        inels, '_InelsBus3__writeValues', return_value=None
                    ) as write:
                        inels.write(lights[0], 0)
                        self.assertEqual(write.call_count, 1)

    def test_InelsDevice_loadFromJson(self):
        """Testing right assigments of all properties of objects."""

        proxy = InelsBus3(TEST_HOST, TEST_PORT)

        # THERM
        therm = TEST_DATA_THERM
        therm_dev = InelsDevice(
            therm['name'], therm['therm'], DeviceType.THERM, proxy)
        therm_dev.loadFromJson(therm)

        self.assertEqual(therm['name'], therm_dev.title)
        self.assertIsNotNone(therm_dev.id)
        self.assertEqual(therm['rele'], therm_dev.rele)
        self.assertEqual(therm['therm'], therm_dev.temp_current)
        self.assertEqual(therm['stateth'], therm_dev.temp_set)
        self.assertEqual(False, therm_dev.read_only)
        self.assertEqual(DeviceType.THERM, therm_dev.type)

        # LIGHT
        therm = TEST_DATA_LIGHT
        therm_dev = InelsDevice(
            therm['name'], therm['id'], DeviceType.LIGHT, proxy)
        therm_dev.loadFromJson(therm)

        self.assertEqual(therm['name'], therm_dev.title)
        self.assertEqual(therm['id'], therm_dev.id)
        self.assertEqual(False, therm_dev.read_only)
        self.assertEqual(DeviceType.LIGHT, therm_dev.type)
        self.assertIsNone(therm_dev.rele)

        # SWITCH
        therm = TEST_DATA_SWITCH
        therm_dev = InelsDevice(
            therm['name'], therm['id'], DeviceType.SWITCH, proxy)
        therm_dev.loadFromJson(therm)

        self.assertEqual(therm['name'], therm_dev.title)
        self.assertEqual(therm['id'], therm_dev.id)
        self.assertEqual(False, therm_dev.read_only)
        self.assertEqual(DeviceType.SWITCH, therm_dev.type)
        self.assertIsNone(therm_dev.temp_current)

        # GARAGE
        therm = TEST_DATA_GARAGE
        therm_dev = InelsDevice(
            therm['name'], therm['id'], DeviceType.GARAGE, proxy)
        therm_dev.loadFromJson(therm)

        self.assertEqual(therm['name'], therm_dev.title)
        self.assertEqual(therm['id'], therm_dev.id)
        self.assertEqual(False, therm_dev.read_only)
        self.assertEqual(DeviceType.GARAGE, therm_dev.type)
        self.assertIsNone(therm_dev.temp_current)

        # HEATING
        therm = TEST_DATA_HEATING
        therm_dev = InelsDevice(
            therm['name'], therm['id'], DeviceType.HEATING, proxy)
        therm_dev.loadFromJson(therm)

        self.assertEqual(therm['name'], therm_dev.title)
        self.assertEqual(therm['id'], therm_dev.id)
        self.assertEqual(False, therm_dev.read_only)
        self.assertEqual(DeviceType.HEATING, therm_dev.type)
        self.assertIsNone(therm_dev.temp_current)

        # SHUTTER
        therm = TEST_DATA_SHUTTER
        therm_dev = InelsDevice(
            therm['name'], therm['up'], DeviceType.SHUTTER, proxy)
        therm_dev.loadFromJson(therm)

        self.assertEqual(therm['name'], therm_dev.title)
        self.assertEqual(False, therm_dev.read_only)
        self.assertEqual(therm['up'], therm_dev.up)
        self.assertEqual(therm['down'], therm_dev.down)
        self.assertEqual(DeviceType.SHUTTER, therm_dev.type)
        self.assertIsNotNone(therm_dev.id)



"""Inels switch class for iNels BUS."""

from inels.const import (
    ATTR_SWITCH_ON
)


class pySwitch:
    """Switch class based on InelsDevice."""

    def __init__(self, device):
        """Initialize of object InelsSwitch."""
        self.device = device

    @property
    def state(self):
        """Return the state of the switch."""
        if (self.device.value is None):
            self.device.observe()

        val = str(self.device.value)
        attr = ATTR_SWITCH_ON

        # it is expression of ternary operator
        return (True if val == attr else False)

    def toggle(self):
        """Toogle the state of the switch."""
        new_state = True if self.state is False else False

        self.device.set_value(1 if new_state is True else 0)



"""Unit testing of iNels switch library."""

from tests.const_test import (
    TEST_DATA_SWITCH,
    TEST_HOST,
    TEST_INELS_BUS3_NAMESPACE,
    TEST_PORT
)

from inels.cu3 import InelsBus3
from inels.device import (
    InelsDevice,
    DeviceType
)

from inels.device.pySwitch import pySwitch

from unittest.mock import patch
from unittest import TestCase


class pySwitchTest(TestCase):
    """Class to test iNels switch library."""

    def setUp(self):
        """Setup all necesary instances and mocks."""
        self.patches = [
            patch(f'{TEST_INELS_BUS3_NAMESPACE}.ping', return_value=True),
            patch(f'{TEST_INELS_BUS3_NAMESPACE}.read',
                  return_value={TEST_DATA_SWITCH['id']: 1}),
            patch('inels.device.InelsDevice._write', return_value=None)
        ]

        self.proxy = InelsBus3(TEST_HOST, TEST_PORT)
        self.device = InelsDevice(TEST_DATA_SWITCH['name'],
                                  TEST_DATA_SWITCH['id'],
                                  DeviceType.SWITCH, self.proxy)

        self.switch = pySwitch(self.device)
        # patching ping method in IneslBus3. It will be executed every test
        for p in self.patches:
            p.start()

    def tearDown(self):
        """Destroy all instances and mocks."""
        self.proxy = None
        self.device = None
        self.switch = None
        patch.stopall()

    def test_initialize_switch(self):
        """Initialize switch and test state."""
        self.assertEqual(TEST_DATA_SWITCH['id'], self.switch.device.id)

    def test_state_switch(self):
        """Test state of the switch."""
        state = self.switch.state
        self.assertEqual(state, True)

    def test_toggle_switch(self):
        """Toogle of the switch."""
        self.switch.toggle()
        self.assertEqual(self.switch.state, False)

        self.switch.toggle()
        self.assertEqual(self.switch.state, True)



"""Library specified for iNels BUS CU3."""
import logging

from dataclasses import dataclass
from inels.device import DeviceType, InelsDevice
from xmlrpc.client import ServerProxy

_LOGGER = logging.getLogger(__name__)


class InelsBus3:
    """Class for iNels BUS CU3 version."""

    def __init__(self, host, port):
        """Initialize InelsBus3 class."""
        self.__host = host
        self.__port = port
        self.__proxy = None

    @property
    def proxy(self):
        """Proxy of the cu3 server."""
        if isinstance(self._InelsBus3__proxy, ServerProxy):
            return self.__proxy

        self.__proxy = self.__conn()
        self.ping()
        return self.__proxy

    def __conn(self):
        """Instantient InelsBus3 class."""
        try:
            con = ServerProxy(self.__host + ":" + str(self.__port))
            return con
        except BlockingIOError as err:
            raise InelsBusConnectionException(err.errno, err.strerror, err)
        except Exception as err:
            raise InelsBusException(
                "common_exception", "Exception occur", err)

    def ping(self):
        """Check connection iNels BUS with ping."""
        return self.proxy.ping()

    def getPlcIp(self):
        """Get Ip address of PLC."""
        return self.proxy.getPlcIP()

    def getRooms(self):
        """List of all rooms from Connection server website."""
        return self.proxy.getRooms()

    def getRoomDevicesRaw(self, room_name):
        """List of all devices in deffined room."""
        return self.proxy.getRoomDevices(room_name)

    def getRoomDevices(self, room_name):
        """List of all devices in defined room."""
        return self.__roomDevicesToJson(room_name)

    def getAllDevices(self):
        """Get all devices from all rooms."""
        rooms = self.getRooms()

        devices = []
        for room in rooms:
            devices += self.getRoomDevices(room)

        return devices

    def read(self, device_ids):
        """Get the value from the proxy by device id."""
        if not isinstance(device_ids, list):
            raise InelsBusDataTypeException(
                'readDeviceData', f'{device_ids} is not a list!')
        return self.__readDeviceData(device_ids)

    def write(self, device, value):
        """Write data to multiple devices."""
        if not isinstance(device, InelsDevice):
            raise InelsBusDataTypeException(
                'readDeviceData', f'{device} is not object InelsDevice')
        try:
            obj = {}
            obj[device.id] = value

            self.__writeValues(obj)
        except Exception as err:
            raise InelsBusException("write_proxy", err)

    def __writeValues(self, command):
        """Write data to the proxy."""
        self.proxy.writeValues(command)

    def __roomDevicesToJson(self, room_name):
        """Create json object from devices listed in preffered room."""
        d_type = None
        devices = []

        raw_list = self.getRoomDevicesRaw(room_name)
        list = raw_list.split('\n')

        for item in list:
            start = len(item) - 1
            end = len(item)
            if start > 0:
                if item[start:end] == ":":
                    d_type = item[0:start]
                else:
                    json_dev = item.split('" ')
                    obj = {}
                    obj["group"] = room_name

                    for prop in json_dev:
                        frag = prop.split("=")
                        obj[frag[0]] = frag[1].replace("\"", " ").strip()

                    device = self.__loadDevice(obj, d_type)
                    devices.append(device)

        return devices

    def __readDeviceData(self, device_names):
        """Reading devices data from proxy."""
        return self.proxy.read(device_names)

    def __loadDevice(self, json_device, d_type: str):
        """Load device to the object structure"""
        id = None
        if 'inels' in json_device:
            id = json_device['inels']

        device = InelsDevice(
            json_device['name'],
            id,
            DeviceType.is_in(d_type),
            self.proxy)
        # load data from json
        device.loadFromJson(json_device)

        return device

    @staticmethod
    def __get_val_name(device: InelsDevice, d_type: DeviceType):
        """SHUTTER is quite specific object. It does not have ID.
        UP and DOWN is only defined, which each means ID.
        I need to get a list of device.id to send them to the proxy method
        read. When SHUTTER comes, than I need to make a
        list of up and down props. And send them together to the proxy."""
        return {
            DeviceType.SHUTTER: [device.up, device.down],
            DeviceType.THERM: [device.temp_current, device.temp_set]
        }.get(d_type, [device.id])


@dataclass
class InelsBusException(Exception):
    """Base iNels BUS exception."""
    code: str
    message: str
    trace: Exception = None


@dataclass
class InelsBusConnectionException(InelsBusException):
    """Connection exception class."""


@dataclass
class InelsBusDataTypeException(InelsBusException):
    """Bad type exception."""



"""Unit testing of iNels device library."""

from tests.const_test import (
    TEST_DATA_SWITCH,
    TEST_HOST,
    TEST_INELS_BUS3_NAMESPACE,
    TEST_PORT
)

from inels.cu3 import InelsBus3
from inels.device import InelsDevice, DeviceType

from unittest.mock import patch
from unittest import TestCase


class DeviceTest(TestCase):
    """Class to test iNels device library."""

    def setUp(self):
        """Setup all necesary instances and mocks."""
        self.patches = [
            patch(f'{TEST_INELS_BUS3_NAMESPACE}.ping', return_value=True),
            patch(f'{TEST_INELS_BUS3_NAMESPACE}.read',
                  return_value={TEST_DATA_SWITCH['id']: 1}),
            patch('inels.device.InelsDevice._write', return_value=None)
        ]

        self.proxy = InelsBus3(TEST_HOST, TEST_PORT)
        self.device = InelsDevice(TEST_DATA_SWITCH['name'],
                                  TEST_DATA_SWITCH['id'],
                                  DeviceType.SWITCH, self.proxy)

        # patching ping method in IneslBus3. It will be executed every test
        for p in self.patches:
            p.start()

    def tearDown(self):
        """Destroy all instances and mocks."""
        self.proxy = None
        self.device = None
        patch.stopall()

    def test_initialize_device(self):
        """Class instance test."""
        self.assertEqual(True, self.proxy.ping())

        self.device.loadFromJson(TEST_DATA_SWITCH)

        self.assertEqual(TEST_DATA_SWITCH['name'], self.device.title)
        self.assertEqual(TEST_DATA_SWITCH['id'], self.device.id)
        self.assertEqual(self.device.value, None)

    def test_observer_device(self):
        """Test the observer value of the device."""
        dev_value = self.device.observe()
        self.assertEqual(1, dev_value)
        self.assertEqual(1, self.device.value)

    def test_set_value_device(self):
        """Test setting the value of the device."""
        self.device.set_value(0)
        self.assertEqual(0, self.device.value)



"""Const of iNels BUS."""

ATTR_SWITCH_OFF = "0"
ATTR_SWITCH_ON = "1"

NAME = "pyinels"

RANGE_BLIND = (0, 100)
RANGE_BRIGHTNESS = (0.0, 255.0)

SUPPORT_BRIGHTNESS = 1
SUPPORT_COLOR_TEMP = 4
SUPPORT_HEX_COLOR = 8
SUPPORT_RGB_COLOR = 2

VERSION = "0.2.1"



"""Const of iNels BUS tests."""
TEST_HOST = "http://localhost"

TEST_INELS_NAMESPACE = "inels.cu3"
TEST_INELS_BUS3_NAMESPACE = "inels.cu3.InelsBus3"

TEST_PORT = "1000"
TEST_RAW_DEVICES = """garage:
name="Doors" column="0" inels="Doors_Garage" read_only="no" row="1"
lights:
name="Main light" column="0" inels="SV_12_Garage" read_only="no" row="0"
name="Wall light" column="1" inels="SV_Wall_Garage" read_only="no" row="0"
"""
TEST_ROOMS = ['Basement', 'First floor', 'Front yard',
              'Back yard', 'Attic', 'Garrage']

TEST_DATA_THERM = {
    'name': 'Valve in bedroom',
    'therm': 'Bedroom_temerature',
    'rele': 'DIGITALGROUP_0001',
    'read_only': 'no',
    'stateth': 'Bedroom_stateth'
}

TEST_DATA_LIGHT = {
    'name': 'Kitchen main light',
    'id': 'KITCHEN_MAIN_LIGHT',
    'read_only': 'no'
}

TEST_DATA_SWITCH = {
    'name': 'Kitchen kettel switch',
    'id': 'KITCHEN_KETTLE_SWITCH',
    'read_only': 'no'
}

TEST_DATA_SHUTTER = {
    'name': 'Kitchen shuuter',
    'up': 'KITCHEN_SHUUTER_1_UP',
    'down': 'KITCHEN_SHUUTER_1_DOWN',
    'read_only': 'no'
}

TEST_DATA_GARAGE = {
    'name': 'Garage door',
    'id': 'GARAGE_DOOR',
    'read_only': 'no'
}

TEST_DATA_HEATING = {
    'name': 'Holiday',
    'id': 'HOLIDAY_HEATING',
    'read_only': 'no'
}



"""Unit testing of iNels light library."""

from tests.const_test import (
    TEST_DATA_LIGHT,
    TEST_HOST,
    TEST_INELS_BUS3_NAMESPACE,
    TEST_PORT
)

from inels.cu3 import InelsBus3
from inels.device import (
    InelsDevice,
    DeviceType
)

from inels.device.pyLight import pyLight

from unittest.mock import patch
from unittest import TestCase


class pyLightTest(TestCase):
    """Class to test iNels light library."""

    def setUp(self):
        """Setup all necesary instances and mocks."""
        self.patches = [
            patch(f'{TEST_INELS_BUS3_NAMESPACE}.ping', return_value=True),
            patch(f'{TEST_INELS_BUS3_NAMESPACE}.read',
                  return_value={TEST_DATA_LIGHT['id']: 1}),
            patch('inels.device.InelsDevice._write', return_value=None)
        ]

        proxy = InelsBus3(TEST_HOST, TEST_PORT)
        device = InelsDevice(TEST_DATA_LIGHT['name'],
                             TEST_DATA_LIGHT['id'],
                             DeviceType.LIGHT, proxy)

        self.light = pyLight(device)
        # patching ping method in IneslBus3. It will be executed every test
        for p in self.patches:
            p.start()

    def tearDown(self):
        """Destroy all instances and mocks."""
        self.light = None
        patch.stopall()

    def test_initialize_light(self):
        """Initialize light test."""
        self.assertEqual(TEST_DATA_LIGHT['id'], self.light.device.id)

    def test_state_light(self):
        """Test state of the light."""
        state = self.light.state
        self.assertEqual(state, True)

    def test_has_not_brightness(self):
        """Test that the light has no brightness feature."""
        brightness_support = self.light.has_brightness
        self.assertFalse(brightness_support)

    def test_has_brightness(self):
        """Test that the light has brightness feature."""
        self.light.device.set_value(25.0)
        brightness_support = self.light.has_brightness

        self.assertTrue(brightness_support)

    def test_set_brightness(self):
        """Test set brightness of the light."""
        value = self.light.device.observe()
        # at the begining the light has no brightness support

        self.light.set_brightness(0.0)
        self.assertFalse(self.light.has_brightness)
        # no change
        self.assertEqual(value, self.light.device.value)

        # set brightness support to the light
        self.light.device.set_value(0.0)
        self.assertTrue(self.light.has_brightness)

        self.light.set_brightness(25.0)
        self.assertNotEqual(value, self.light.device.value)
        self.assertEqual(self.light.device.value, 25.0)

        self.light.set_brightness(256.0)
        # brightness is out of range, then no changes
        self.assertNotEqual(self.light.device.value, 256.0)
        self.assertEqual(self.light.device.value, 25.0)

        # remove brightness support
        self.light.device.set_value(0)
        self.assertFalse(self.light.has_brightness)
        self.light.set_brightness(25.0)
        self.assertNotEqual(self.light.device.value, 25.0)
        self.assertEqual(self.light.device.value, 0)

    def test_set_state(self):
        """Test set state of the light."""
        # set the light to turn off without brightness support
        # for testing purposes
        self.light.device.set_value(0)

        self.light.set_state()
        self.assertTrue(self.light.state)
        self.assertEqual(self.light.device.value, 1)

        # turn off the ligth
        self.light.set_state()
        self.assertFalse(self.light.state)
        self.assertEqual(self.light.device.value, 0)

        # set the brightness support
        self.light.device.set_value(0.0)
        # turn on the ligt wit brightness
        self.light.set_state()
        self.assertTrue(self.light.state)
        self.assertEqual(self.light.device.value, 255.0)
